# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from . import utils
from dataclasses import dataclass
from .base import Base
from functools import cached_property
from pathlib import Path
import re, time
import rich.box
from rich.text import Text
from rich.table import Table


class Device(Base):
    """
    Accessors for the state of ARM Cortex-M CPU, STM32 identifiers, uptime, and
    NuttX kernel internals.
    """
    _HRT_CNT = 0x4001_0424
    _SCS_ICTR = 0xE000_E004
    _SCB_CPUID = 0xE000_ED00
    _SCB_VTOR = 0xE000_ED08
    _SCS_SHPR = 0xE000_ED18
    _NVIC_ISER = 0xE000_E100
    _NVIC_ISPR = 0xE000_E200
    _NVIC_IABR = 0xE000_E300
    _NVIC_IPR = 0xE000_E400

    # Try them all until one of them does not return zero
    _DBG_IDCODE = [0xE004_2000, 0x5C00_1000]

    @cached_property
    def _IDCODE_REVISION(self):
        return {
            0x1000: "A",
            0x1001: "Z",
            0x1003: "Y",
            0x1007: "4",
            0x2001: "X",
            0x2003: "Y",
        }.get(self.rev, "")

    @cached_property
    def _IDCODE_DEVICE(self):
        return {
            0x0415: "STM32L47/L48xx",
            0x0451: "STM32F76xx, STM32F77xx",
            0x0450: "STM32H742, STM32H743/753, STM32H750",
            0x0483: "STM32H723/733, STM32H725/735, STM32H730",
        }.get(self.devid, "Unknown")

    @cached_property
    def _GPIOS(self):
        return {
            0x0415: list(range(0x4800_0000, 0x4800_2000, 0x400)),
            # FMUv5x/v6x don't use ports J and K
            0x0451: list(range(0x4002_0000, 0x4002_2001, 0x400)),
            0x0450: list(range(0x5802_0000, 0x5802_2001, 0x400)),
            0x0483: list(range(0x5802_0000, 0x5802_2C01, 0x400)),
        }.get(self.devid, [])

    @cached_property
    def _SYSTEM_MEMORIES(self):
        mems = [(0xE000_0000, 0x10_0000)]
        mems += {
            0x0415: [(0x1FFF_7500, 0x100)],
            0x0451: [(0x1FF0_F420, 0x100), (0x1FFF_7B00, 0x100)],
            **dict.fromkeys([0x0450, 0x0483],
                    [(0x1FF1_E800, 0x100), (0x58000524, 4)])
        }.get(self.devid, [])
        return mems

    @cached_property
    def _PERIPHERALS(self):
        if self._SVD_FILE is None: return []
        from cmsis_svd.parser import SVDParser
        device = SVDParser.for_xml_file(self._SVD_FILE).get_device()
        registers = [(per.base_address + r.address_offset,
                      per.base_address + r.address_offset + r.size//8)
                     for per in device.peripherals for r in per.registers]
        registers.sort()
        ranges = []
        cluster = 512
        current = registers[0]
        for (regl, regh) in registers[1:]:
            if (current[1] + cluster) > regh:
                current = (current[0], regh)
            else:
                ranges.append(current)
                current = (regl, regh)
        ranges.append(current)
        return [(r[0], r[1] - r[0]) for r in ranges]

    @cached_property
    def _MEMORIES(self):
        mems = {
            0x0415: [
                (0x1000_0000, 0x20000), # SRAM2
                (0x2000_0000, 0x20000), # SRAM1
            ],
            0x0451: [
                # (0x0000_0000, 0x04000), # ITCM
                (0x2000_0000, 0x80000), # DTCM, SRAM1, SRAM2
            ],
            **dict.fromkeys([0x0450, 0x0483], [
                # (0x0000_0000, 0x10000), # ITCM
                (0x2000_0000, 0x20000), # DTCM
                (0x2400_0000, 0x80000), # AXI_SRAM
                (0x3000_0000, 0x48000), # SRAM1, SRAM2, SRAM3
                (0x3800_0000, 0x10000), # SRAM4
                (0x3880_0000, 0x01000), # Backup SRAM
                (0x5C00_1000, 4),       # IDCODE
            ]),
        }.get(self.devid, [])
        mems += self._PERIPHERALS
        mems += self._SYSTEM_MEMORIES
        return sorted(mems)

    @cached_property
    def _SVD_FILE(self):
        return {
            # 0x0415: Path(__file__).parents[1] / "data/STM32L4x6.svd",
            0x0451: Path(__file__).parents[1] / "data/STM32F765.svd",
            0x0450: Path(__file__).parents[1] / "data/STM32H753.svd",
            # 0x0483: Path(__file__).parents[1] / "data/STM32H7x3.svd",
        }.get(self.devid)

    @dataclass
    class Gpio:
        port: str
        index: int
        moder: int
        otyper: int
        speedr: int
        pupdr: int
        idr: int
        odr: int
        lockr: int
        afr: int

    @dataclass
    class Irq:
        index: int
        """Index of the IRQ starting at zero"""
        handler: "gdb.Block"
        """Function handler of the IRQ"""
        priority: int
        """Priority of the IRQ"""
        is_enabled: bool
        """Is interrupt enabled"""
        is_pending: bool
        """Is interrupt pending"""
        is_active: bool
        """Is interrupt active"""

    @dataclass
    class IrqNuttX(Irq):
        arg: "gdb.Value"
        """Optional argument passed to the IRQ handler"""

    def __init__(self, gdb):
        super().__init__(gdb)
        self.architecture = self._arch.name()
        self._hrt_counter = self.addr_ptr(self._HRT_CNT, "uint16_t")

    @cached_property
    def _hrt_base(self):
        return self.lookup_static_symbol_in_function_ptr("base_time", "hrt_absolute_time")

    def _read_bits(self, base, index, size):
        shift = ((index * size) % 64)
        offset = (index * size) // 64
        mask = ((1 << size) - 1) << shift
        bits = self.read_memory(base + offset, 8).cast("Q")[0]
        return (bits & mask) >> shift

    def _priority(self, index):
        if index == -15: return -3
        if index == -14: return -2
        if index == -13: return -1
        if index < 0: return self._read_bits(self._SCS_SHPR, index + 16, 8)
        return self._read_bits(self._NVIC_IPR, index, 8)

    def _enabled(self, index):
        # if index < 0: return self._read_bits1(self._SCS_, index);
        return self._read_bits(self._NVIC_ISER, index, 1)

    def _pending(self, index):
        return self._read_bits(self._NVIC_ISPR, index, 1)

    def _active(self, index):
        return self._read_bits(self._NVIC_IABR, index, 1)

    @cached_property
    def uptime(self) -> int:
        """The uptime in microseconds as read from the NuttX high-resolution timer (HRT)"""
        if self._hrt_base is None: return 0
        return int(self._hrt_base.dereference() + self._hrt_counter.dereference())

    @cached_property
    def max_interrupts(self) -> int:
        """Maximum number of interrupts implemented on this device"""
        return 32 * (self.read_uint(self._SCS_ICTR, 4) + 1)

    @cached_property
    def vector_table(self) -> list[Irq]:
        """The vector table as stored inside the SCB->VTOR"""
        # SCS = 0xE000E000, SCB = 0xE000ED00, SCB->VTOR = 0xE000ED08
        # TODO: Cortex-M0 has no VTOR, Cortex-M0+ has one if implemented
        vtor = self._gdb.Value(self._SCB_VTOR).cast(self.uint32.pointer())
        vtor = vtor.dereference().cast(self.uint32.pointer())
        entries = []
        for ii in range(self.max_interrupts):
            ii -= 16
            entries.append(self.Irq(ii, self.block(vtor[ii+16]), self._priority(ii),
                                    self._enabled(ii), self._pending(ii), self._active(ii)))
        return entries

    @cached_property
    def vector_table_nuttx(self) -> dict[int, IrqNuttX]:
        """
        The index, handler and argument of the NuttX interrupt table if the
        handler is not empty and is not `irq_unexpected_isr`.
        """
        g_irqvector = self._gdb.lookup_global_symbol("g_irqvector")
        vectors = {}
        for ii, vector in enumerate(utils.gdb_iter(g_irqvector)):
            if handler := vector["handler"]:
                block = self.block(handler)
                if block.function and block.function.name == "irq_unexpected_isr":
                    continue
                ii -= 16
                vectors[ii] = self.IrqNuttX(ii, block, self._priority(ii),
                                            self._enabled(ii), self._pending(ii),
                                            self._active(ii), vector["arg"])
        return vectors

    @property
    def gpios(self) -> list[Gpio]:
        """List of GPIOs as read from the device."""
        result = []
        for jj, gper_addr in enumerate(self._GPIOS):
            port = chr(ord("A") + jj)
            gper = self.read_memory(gper_addr, 0x28).cast("I")
            for ii in range(16):
                _1v = lambda index: (gper[index] >> ii) & 0x1
                _2v = lambda index: (gper[index] >> ii*2) & 0x3
                moder, speedr, pupdr = _2v(0), _2v(2), _2v(3)
                otyper, idr, odr, lockr = _1v(1), _1v(4), _1v(5), _1v(7)
                afr = ((gper[9] << 32 | gper[8]) >> ii*4) & 0xf
                g = self.Gpio(port, ii, moder, otyper, speedr, pupdr, idr, odr, lockr, afr)
                result.append(g)
        return result

    def coredump(self, memories: list[tuple[int, int]] = None, with_flash: bool = False) -> tuple[str, int]:
        """
        Reads the memories and registers and returns them as a formatted string
        that is compatible with CrashDebug (see `emdbg.debug.crashdebug`).

        :param memories: list of memory ranges (start, size) to dump
        :param with_flash: also dump the entire non-volatile storage
        :return: coredump formatted as string and coredump size
        """
        if memories is None:
            memories = self._MEMORIES
        if with_flash and self.flash_size:
            memories += [(0x0800_0000, self.flash_size)]
        lines = []
        total_size = 0
        for addr, size in memories:
            try:
                data = self.read_memory(addr, size).cast("I")
                total_size += size
            except Exception as e:
                print(f"Failed to read whole range [{addr:#x}, {addr+size:#x}]! {e}")
                data = []
                for offset in range(0, size, 4):
                    try:
                        data.append(self.read_memory(addr + offset, 4).cast("I")[0])
                        total_size += 4
                    except Exception as e:
                        print(f"Failed to read uint32_t {addr+offset:#x}! {e}")
                        data.append(0)
                        continue
            for ii, values in enumerate(utils.chunks(data, 4, 0)):
                values = (hex(v & 0xffffffff) for v in values)
                lines.append(f"{hex(addr + ii * 16)}: {' '.join(values)}")

        for name, value in self.registers.items():
            if re.match(r"d\d+", name):
                lines.append(f"{name:<28} {float(value):<28} (raw {value & 0xffffffffffffffff:#x})")
            elif re.match(r"s\d+", name):
                lines.append(f"{name:<28} {float(value):<28} (raw {value & 0xffffffff:#x})")
            else:
                lines.append(f"{name:<28} {hex(value & 0xffffffff):<28} {int(value)}")

        return "\n".join(lines), total_size

    @cached_property
    def cpuid(self) -> int:
        """The SCB->CPUID value"""
        return self.read_uint(self._SCB_CPUID, 4)

    @cached_property
    def idcode(self) -> int:
        """The STM32-specific DBG->IDCODE value"""
        for addr in self._DBG_IDCODE:
            if idcode := (self.read_uint(addr, 4) & 0xffff0fff):
                return idcode
        return 0

    @cached_property
    def flash_size(self) -> int:
        """The FLASH size in bytes"""
        addr = {
            0x0415: 0x1FFF_75E0,
            0x0451: 0x1FF0_F442,
            0x0450: 0x1FF1_E880,
            0x0483: 0x1FF1_E880,
        }.get(self.devid)
        if addr is None: return 0
        return self.read_uint(addr, 2) * 1024

    @cached_property
    def line(self) -> str:
        """The device family and name"""
        if self.devid == 0x0483:
            return self.read_string(0x1FF1_E8C0, length=4)[::-1]
        return self._IDCODE_DEVICE

    @cached_property
    def uid(self) -> int:
        """The device's unique identifier as a big integer"""
        addr = {
            0x0415: 0x1FFF_7590,
            0x0451: 0x1FF0_F420,
            0x0450: 0x1FF1_E800,
            0x0483: 0x1FF1_E800,
        }.get(self.devid)
        if addr is None: return 0
        return int.from_bytes(self.read_memory(addr, 3*4).tobytes(), "little")

    @cached_property
    def package(self) -> str:
        """The device package"""

        if self.devid == 0x0415:
            return {
                0b00000: "LQFP64",
                0b00010: "LQFP100",
                0b00011: "UFBGA132",
                0b00100: "LQFP144, UFBGA144, WLCSP72, WLCSP81 or WLCSP86",
                0b10000: "UFBGA169, WLCSP115",
                0b10001: "WLCSP100",
            }.get(self.read_uint(0x1FFF_7500, 1) & 0x1f, "Reserved")
        if self.devid == 0x0451:
            return {
                0b111: "LQFP208 or TFBGA216",
                0b110: "LQFP208 or TFBGA216",
                0b101: "LQFP176",
                0b100: "LQFP176",
                0b011: "WLCSP180",
                0b010: "LQFP144 ",
                0b001: "LQFP100",
            }.get(self.read_uint(0x1FFF_7BF1, 1) & 0x7, "Reserved")
        if self.devid == 0x0450:
            return {
                0b0000: "LQFP100",
                0b0010: "TQFP144",
                0b0101: "TQFP176/UFBGA176",
                0b1000: "LQFP208/TFBGA240",
            }.get(self.read_uint(0x58000524, 1) & 0xf, "All pads enabled")
        if self.devid == 0x0483:
            return {
                0b0000: "VFQFPN68 Industrial",
                0b0001: "LQFP100 Legacy / TFBGA100 Legacy",
                0b0010: "LQFP100 Industrial",
                0b0011: "TFBGA100 Industrial",
                0b0100: "WLCSP115 Industrial",
                0b0101: "LQFP144 Legacy",
                0b0110: "UFBGA144 Legacy",
                0b0111: "LQFP144 Industrial",
                0b1000: "UFBGA169 Industrial",
                0b1001: "UFBGA176+25 Industrial",
                0b1010: "LQFP176 Industrial",
            }.get(self.read_uint(0x58000524, 1) & 0xf, "All pads enabled")
        return "?"

    @cached_property
    def devid(self) -> int:
        """The STM32-specific device id part of DBG->IDCODE"""
        return self.idcode & 0xfff

    @cached_property
    def rev(self) -> int:
        """The STM32-specific revision part of DBG->IDCODE"""
        return self.idcode >> 16


def discover(gdb) -> Table:
    """
    Reads the device identifier registers and outputs a human readable string if possible.
    :return: description string
    """
    dev = Device(gdb)
    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("Device")
    table.add_column("Revision")
    table.add_column("Flash")
    table.add_column("Package")
    table.add_column("UID")
    table.add_row(f"{dev.devid:#4x}: {dev.line}",
                  f"{dev.rev:#4x}: rev {dev._IDCODE_REVISION}",
                  f"{dev.flash_size//1024}kB", dev.package,
                  f"{dev.uid:x}")
    return table


def all_registers(gdb) -> dict[str, int]:
    """Return a dictionary of register name and values"""
    return Device(gdb).registers


def all_registers_as_table(gdb, columns: int = 3) -> Table:
    """
    Format the Cortex-M CPU+FPU registers and their values into a simple table.

    :param columns: The number of columns to spread the registers across.
    """
    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("Name")
    table.add_column("Hexadecimal", justify="right")
    table.add_column("Decimal", justify="right")
    table.add_column("Binary", justify="right")
    for reg, value in all_registers(gdb).items():
        table.add_row(reg, f"{value:x}", str(value), f"{value:b}")
    return table


def vector_table(gdb) -> dict[int, Device.IrqNuttX]:
    """Return a dictionary of NuttX interrupt numbers and their handlers with arguments"""
    return Device(gdb).vector_table_nuttx


def vector_table_as_table(gdb, columns: int = 1) -> Table:
    """
    Format the NuttX interrupts and their handlers with arguments into a simple table.

    :param columns: The number of columns to spread the interrupts across.
    """
    def _fname(block):
        count = 0
        while block and block.function is None:
            block = block.superblock
            count += 1
        if block.function:
            return f"{block.function.name} {'^' * count}"
        return f"{block} ?"

    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("IRQ", justify="right")
    table.add_column("EPA")
    table.add_column("Prio")
    table.add_column("Address")
    table.add_column("Function")
    table.add_column("Argument")
    for idx, irq in vector_table(gdb).items():
        table.add_row(str(idx),
                      ("e" if irq.is_enabled else " ") +
                      ("p" if irq.is_pending else " ") +
                      ("a" if irq.is_active else " "),
                      f"{irq.priority:x}", hex(irq.handler.start),
                      _fname(irq.handler), str(irq.arg) or "",
                      style="bold blue" if irq.is_active else None)
    return table


def coredump(gdb, memories: list[tuple[int, int]] = None,
             with_flash: bool = False, filename: Path = None):
    """
    Dumps the memories and register state into a file.

    :param memories: List of (addr, size) tuples that describe which memories to dump.
    :param with_flash: Also dump the entire non-volatile storage.
    :param filename: Target filename, or `coredump_{datetime}.txt` by default.
    """
    if filename is None:
        filename = utils.add_datetime("coredump.txt")
    print("Starting coredump...", flush=True)
    start = time.perf_counter()
    output, size = Device(gdb).coredump(memories, with_flash)
    Path(filename).write_text(output)
    end = time.perf_counter()
    print(f"Dumped {size//1000}kB in {(end - start):.1f}s ({int(size/((end - start)*1000))}kB/s)")


def all_gpios_as_table(gdb, pinout: dict[str, tuple[str, str]] = None,
                       fn_filter = None, sort_by = None, columns: int = 2) -> Table:
    """
    Reads the GPIO peripheral space and prints a table of the individual pin
    configuration, input/output state and alternate function. If a pinout is
    provided, the pins will be matched with their names and functions.

    Config: Condensed view with omitted defaults.
        MODER:  IN=Input, OUT=Output, ALT=Alternate Function, AN=Analog,
        OTYPER: +OD=OpenDrain, (+PP=PushPull omitted),
        PUPDR:  +PU=PullUp, +PD=PullDown, (+FL=Floating omitted),
        SPEEDR: +M=Medium, +H=High, +VH=Very High, (+L=Low omitted).

    Input (IDR), Output (ODR): _=Low, ^=High
        Input only shown for IN, OUT, and ALT.
        Output only shown for OUT.

    Alternate Function (AFR): only shown when config is ALT.
        Consult the datasheet for device-specific mapping.

    :param pinout: A map of port+index -> (name, function).
    :param pin_filter: A filter function that gets passed the entire row as a
                       list and returns `True` if row is accepted.
    :param sort_by: The name of the column to sort by: default is `PIN`.
    :param columns: The number of columns to spread the GPIO table across.
    """
    rows = []
    for gpio in Device(gdb).gpios:
        name = f"{gpio.port}{gpio.index}"
        if pinout is not None:
             list(pinout.get(name, [""]*2))
        config = "".join([["IN", "OUT", "ALT", "AN"][gpio.moder],
                          ["", "+OD"][gpio.otyper],
                          ["", "+PU", "+PD", "+!!"][gpio.pupdr],
                          ["", "+M", "+H", "+VH"][gpio.speedr],
                          ["", "+L"][gpio.lockr]])
        idr = "" if gpio.moder == 3 else ["_", "^"][gpio.idr]
        odr = ["_", "^"][gpio.odr] if gpio.moder == 1 else ""
        afr = str(gpio.afr) if gpio.moder == 2 else ""
        row = [name, config, idr, odr, afr]
        if pinout is not None:
            row += list(pinout.get(name, [""]*2))
        if fn_filter is None or fn_filter(row):
            rows.append(row)
    if sort_by is not None:
        def _sort(row):
            pinf = ord(row[0][0]) * 100 + int(row[0][1:])
            if sort_by == 0: return pinf
            if sort_by == 4: return -1 if row[4] == "" else row[4]
            if sort_by == 5: return (row[5], row[6], pinf)
            if sort_by == 6: return (row[6], row[5], pinf)
            return row[sort_by]
        rows.sort(key=_sort)

    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("Pin")
    table.add_column("Config")
    table.add_column("I")
    table.add_column("O")
    table.add_column("AF", justify="right")
    if pinout is not None:
        table.add_column("Name")
        table.add_column("Function")
    for row in rows:
        table.add_row(*row)
    return table




# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
This module provides common analyzers for trace data and debug logs.
"""

from .backtrace import Backtrace, FileSystemBacktrace, SpiBacktrace, I2cBacktrace
from .backtrace import CanBacktrace, UartBacktrace, SemaphoreBacktrace
from .callgraph import callgraph_from_backtrace
from .priority import summarize_semaphore_boostlog
from .hardfault import convert as convert_hardfault



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

import argparse
import emdbg.patch
from inspect import getmembers, isfunction
from pathlib import Path

# Find all patches automatically
map_patches = {n:f for n,f in getmembers(emdbg.patch, isfunction) if not n.startswith("_")}

parser = argparse.ArgumentParser(
    description="Applies")
parser.add_argument(
    "patch",
    choices=map_patches.keys(),
    help="Which patches to apply.")
parser.add_argument(
    "--px4-dir",
    default=".",
    type=Path,
    help="The PX4 root directory you are working on.")
parser.add_argument(
    "-v",
    dest="verbosity",
    action="count",
    help="Verbosity level.")

group = parser.add_mutually_exclusive_group(required=True)
group.add_argument("--apply", action="store_true")
group.add_argument("--restore", action="store_true")

args = parser.parse_args()
emdbg.logger.configure(args.verbosity)

patchset = map_patches[args.patch](args.px4_dir)

if args.apply:
    patchset.do()
if args.restore:
    patchset.undo()



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import re
import math

from datetime import datetime
from itertools import zip_longest
from pathlib import Path


# -----------------------------------------------------------------------------
def gdb_getfield(value: "gdb.Value", name: str, default=None):
    """Find the field of a struct/class by name"""
    for f in value.type.fields():
        if name == f.name:
            return value[name]
    return default


def gdb_iter(obj):
    """yields the values in an array or value with a range"""
    if hasattr(obj, "value"):
        obj = obj.value()
    if hasattr(obj.type, "range"):
        for ii in range(*obj.type.range()):
            yield obj[ii]
    else:
        return []

def gdb_len(obj) -> int:
    """Computes the length of a gdb object"""
    if hasattr(obj.type, "range"):
        start, stop = obj.type.range()
        return stop - start
    else:
        return 1

def gdb_backtrace(gdb) -> str:
    """
    Unfortunately the built-in gdb command `backtrace` often crashes when
    trying to resolve function arguments whose memory is inaccessible due to
    optimizations or whose type is too complex.
    Therefore this is a simpler implementation in Python to avoid GDB crashing.

    ```
    (gdb) px4_backtrace
    #0  0x0800b3be in sched_unlock() at platforms/nuttx/NuttX/nuttx/sched/sched/sched_unlock.c:272
    #1  0x0800b59e in nxsem_post() at platforms/nuttx/NuttX/nuttx/sched/semaphore/sem_post.c:175
    #2  0x0800b5b6 in sem_post() at platforms/nuttx/NuttX/nuttx/sched/semaphore/sem_post.c:220
    #3  0x08171570 in px4::WorkQueue::SignalWorkerThread() at platforms/common/px4_work_queue/WorkQueue.cpp:151
    #4  0x08171570 in px4::WorkQueue::Add(px4::WorkItem*) at platforms/common/px4_work_queue/WorkQueue.cpp:143
    #5  0x0816f8dc in px4::WorkItem::ScheduleNow() at platforms/common/include/px4_platform_common/px4_work_queue/WorkItem.hpp:69
    #6  0x0816f8dc in uORB::SubscriptionCallbackWorkItem::call() at platforms/common/uORB/SubscriptionCallback.hpp:169
    #7  0x0816f8dc in uORB::DeviceNode::write(file*, char const*, unsigned int) at platforms/common/uORB/uORBDeviceNode.cpp:221
    #8  0x0816faca in uORB::DeviceNode::publish(orb_metadata const*, void*, void const*) at platforms/common/uORB/uORBDeviceNode.cpp:295
    #9  0x081700bc in uORB::Manager::orb_publish(orb_metadata const*, void*, void const*) at platforms/common/uORB/uORBManager.cpp:409
    #10 0x0816e6d8 in orb_publish(orb_metadata const*, orb_advert_t, void const*) at platforms/common/uORB/uORBManager.hpp:193
    #11 0x08160912 in uORB::PublicationMulti<sensor_gyro_fifo_s, (unsigned char)4>::publish(sensor_gyro_fifo_s const&) at platforms/common/uORB/PublicationMulti.hpp:92
    #12 0x08160912 in PX4Gyroscope::updateFIFO(sensor_gyro_fifo_s&) at src/lib/drivers/gyroscope/PX4Gyroscope.cpp:149
    #13 0x08040e74 in Bosch::BMI088::Gyroscope::BMI088_Gyroscope::FIFORead(unsigned long long const&, unsigned char) at src/drivers/imu/bosch/bmi088/BMI088_Gyroscope.cpp:447
    #14 0x08041102 in Bosch::BMI088::Gyroscope::BMI088_Gyroscope::RunImpl() at src/drivers/imu/bosch/bmi088/BMI088_Gyroscope.cpp:224
    #15 0x0803fb7a in I2CSPIDriver<BMI088>::Run() at platforms/common/include/px4_platform_common/i2c_spi_buses.h:343
    #16 0x0817164c in px4::WorkQueue::Run() at platforms/common/px4_work_queue/WorkQueue.cpp:187
    #17 0x08171798 in px4::WorkQueueRunner(void*) at platforms/common/px4_work_queue/WorkQueueManager.cpp:236
    #18 0x08014ccc in pthread_startup() at platforms/nuttx/NuttX/nuttx/libs/libc/pthread/pthread_create.c:59
    ```

    :return: the selected frame's backtrace without resolving function argument
    """
    frame = gdb.selected_frame()
    index = 0
    output = []
    while(frame and frame.is_valid()):
        pc = frame.pc()
        if pc > 0xffff_ff00:
            output.append(f"#{index: <2} <signal handler called>")
        else:
            line = "??"
            file = "??"
            if sal := frame.find_sal():
                line = sal.line
                if sal.symtab:
                    file = sal.symtab.fullname()
            if func := frame.function():
                func = func.print_name
                if not func.endswith(")"): func += "()"
            else:
                func = "??"
            output.append(f"#{index: <2} 0x{pc:08x} in {func} at {file}:{line}")
        frame = frame.older()
        index += 1
    return "\n".join(output)

def gdb_relative_location(gdb, location) -> str:
    """
    GDB can only place breakpoint on specific line number inside a file.
    However, if the file changes the line numbers can shift around, which makes
    this method brittle. Therefore this function finds the function inside
    the file and applies a offset or searches for a pattern inside that function
    to determine the correct line number.

    :param location:
        Location `function:+offset` or `function:regex`. In case the function
        uses static linkage you may also need to provide a unique part of the
        filename path to arbitrate multiple identically named static functions
        `file:function:+offset` or `file:function:regex`.

    :return: absolute location string `file:line_number`
    """
    parts = location.split(":")
    file_name = None
    if len(parts) == 3:
        file_name, function_name, line_pattern = parts
    elif len(parts) == 2:
        function_name, line_pattern = parts
    else:
        raise ValueError(f"Unknown location format '{location}'!")()

    function = gdb.lookup_global_symbol(function_name, gdb.SYMBOL_VAR_DOMAIN)
    if function is None:
        # Multiple static symbols may exists, we use the filename to arbitrate
        if functions := gdb.lookup_static_symbols(function_name, gdb.SYMBOL_VAR_DOMAIN):
            file_functions = [(f.symtab.fullname(), f) for f in functions]
            # Arbitrate using file name hint
            if file_name is not None:
                file_functions = [f for f in file_functions if file_name in f[0]]
            if len(file_functions) == 1:
                function = file_functions[0][1]
            else:
                raise ValueError("Multiple functions found:\n - " + "\n - ".join(functions))
    if function is None:
        raise ValueError(f"Cannot find function name '{function_name}'")
    assert function.is_function

    # Find source file and line numbers: how to use file_name?
    file = function.symtab.fullname()
    line_numbers = function.symtab.linetable().source_lines()
    lmin, lmax = min(line_numbers), max(line_numbers)

    if line_pattern.startswith("+"):
        # line offset relative to function
        line = int(line_pattern[1:]) + function.line
    else:
        # regex line pattern, read source file and find the line
        lines = Path(file).read_text().splitlines()
        lines = list(enumerate(lines[lmin:lmax]))
        for ii, line in lines:
            if re.search(line_pattern, line):
                line = lmin + ii
                break
        else:
            lines = "\n  ".join(f"{lmin+l[0]:>4}: {l[1]}" for l in lines)
            raise ValueError(f"Cannot find source line for '{line_pattern}'!\n"
                f"Function '{function_name}' stretches over lines {lmin}-{lmax}.\n")
                # f"Available source lines are:\n  {lines}")

    return f"{file}:{line}"


# -----------------------------------------------------------------------------
def _binary_search(array, value, lo: int, hi: int, direction: int):
    middle = (lo + hi) // 2
    if hi - lo <= 1: return middle

    nslice = [(lo, middle), (middle, hi)]
    pick_lower_half = array[middle] == value
    if direction > 0: pick_lower_half = 1 - pick_lower_half

    lo, hi = nslice[pick_lower_half]
    return _binary_search(array, value, lo, hi, direction)

def binary_search_last(array, value, lo: int = None, hi: int = None):
    """Binary search the last occurrance of value in an array"""
    if lo is None: lo = 0
    if hi is None: hi = len(array)
    return _binary_search(array, value, lo, hi, direction=-1)

def binary_search_first(array, value, lo: int = None, hi: int = None):
    """Binary search the first occurrance of value in an array"""
    if lo is None: lo = 0
    if hi is None: hi = len(array)
    return _binary_search(array, value, lo, hi, direction=1)


# -----------------------------------------------------------------------------
def chunks(iterable, chunk_size: int, fill=None):
    """Convert a iterable into a list of chunks and fill the rest with a value"""
    args = [iter(iterable)] * chunk_size
    return zip_longest(*args, fillvalue=fill)


def add_datetime(filename: str|Path):
    """
    Appends a filename with the current date and time:
    `Year_Month_Day_Hour_Minute_Second`

    Example: `path/name.txt` -> `path/name_2023_04_14_15_03_24.txt`
    """
    filename = Path(filename)
    return filename.with_stem(f"{filename.stem}_{datetime.now().strftime('%Y_%m_%d_%H_%M_%S')}")


# -----------------------------------------------------------------------------
def format_table(fmtstr: str, header: list[str], rows: list[list[str]], columns: int = 1) -> str:
    """
    DEPRECATED: Use `rich.table.Table` instead!

    Formats a list of rows into a table of multiple meta-columns based on the format string.
    Example for formatting an array of registers into a table with three meta-columns:

    ```py
    fmtstr = "{:%d}  {:>%d}  {:>%d}"
    header = ["NAME", "HEX VALUE", "INT VALUE"]
    rows = [[reg, hex(value), value] for reg, value in registers.items()]
    table = utils.format_table(fmtstr, header, rows, 3)
    ```

    :param fmtstr: A string describing the spacing between columns and their
                   alignment. Must have exactly as many entries as the header a
                   rows. Example: two columns, aligned right and left with a
                   brace formatter: `"{:>%d} ({:%d})"` The `%d` is replaced by
                   the max column width by this function.
    :param header: A list of names for the header row. If you specify more than
                   one column, the header will be duplicated.
    :param rows: a list of lists of entries in the table. Each entry will be
                 converted to `str` to count the maximal length of each column.
    :param columns: If a table is very long, the header can be duplicated to the
                    right side of the table to fill more screen space.
    """
    # duplicate and join the format string for each column
    split_horizontal = columns > 0
    columns = abs(columns)
    fmtstr = "      :      ".join([fmtstr] * columns)
    fmtcnt = fmtstr.count("%d")
    column_width = [0] * fmtcnt

    # Interleave the rows for the later chunking
    fill = [""] * (fmtcnt // columns)
    if split_horizontal:
        rows = [val for tup in zip(*chunks(rows, math.ceil(len(rows) / columns), fill)) for val in tup]
    # prepend the duplicated header before the rows
    if header is not None:
        rows = [header] * columns + rows
    # Group the individual rows into multiple columns per row
    rows = [sum(rs, []) for rs in chunks(rows, columns, fill)]

    # collect each line and compute the column width
    lines = []
    for row in rows:
        if len(row) != fmtcnt:
            raise ValueError("Each row have the same number of entries as the format string")
        line = [str(l) for l in row]
        lines.append(line)
        column_width = [max(w, len(l)) for w, l in zip(column_width, line)]

    # Format the format string with the column width first
    fmtstr = fmtstr % tuple(column_width)
    # Now format the actual lines with the formatted format string
    return "\n".join(fmtstr.format(*line).rstrip() for line in lines) + "\n"


# -----------------------------------------------------------------------------
class _Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from pathlib import Path

def read_gdb_log(logfile: Path) -> str:
    """
    Reads a GDB log file and converts a GDB/MI format back into a normal string.
    """
    text = Path(logfile).read_text()
    if '\n~"' in text:
        lines = text.splitlines()
        text = "\n".join(l[2:-1].encode("latin-1").decode('unicode_escape').strip() for l in lines if l.startswith('~"'))
    return text



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import re
from pathlib import Path
from .priority import _BoostOperation

class Frame:
    """
    Parses a GDB frame of a `backtrace` or `px4_backtrace` command.
    """
    def __init__(self, description: str):
        """
        :param description: A single line of the backtrace command starting with a `#`
        """
        self.is_valid = False
        self.description = description
        if match := re.match(r"#(\d+) +(?:0x.+? in )?(.+)\((.*?)\) at (.+?):(\d+)", description):
            self.index = int(match.group(1))
            self.function_name = match.group(2)
            self.args = match.group(3)
            self.filename = match.group(4)
            self.line = match.group(5)
            self.is_valid = True
            self.function = f"{self.function_name}({self.args})"
            self.location = f"{Path(self.filename).relative_to(Path().cwd())}:{self.line}"
            self._unique_location = self.function + "\n" + self.location

    def __hash__(self) -> int:
        return hash(self._node)

    def __eq__(self, other) -> int:
        return self.function == other.function and self.filename == other.filename

    def __repr__(self) -> str:
        return self.function

    def __str__(self) -> str:
        return f"#{self.index:<2} {self.function}"


class Backtrace:
    """
    Holds the entire call chain of a `backtrace` or `px4_backtrace` command.
    """
    EDGES = {}
    COLORS = {}

    def __init__(self, description: str):
        """
        :param description: All lines of the backtrace command
        """
        self.description = description
        tasks = (re.match(r"Task=(.+)", d) for d in description.splitlines())
        tasks = [t.group(1) for t in tasks if t is not None]
        self.task = tasks[0] if tasks else None
        self.type = self.task

        frames = (Frame(d) for d in description.splitlines())
        self.frames = [f for f in frames if f.is_valid]
        self.is_valid = len(self.frames)
        if self.is_valid:
            funcs = {f.function for f in self.frames}
            for name, pattern in self.EDGES.items():
                if any(re.search(pattern, f) for f in funcs):
                    self.type = name
                    break

    def __hash__(self) -> int:
        return hash("".join(f.filename + f.function for f in self.frames))

    def __eq__(self, other) -> int:
        ours = "".join(f.filename + f.function for f in self.frames)
        others = "".join(f.filename + f.function for f in other.frames)
        return ours == others

    def __repr__(self) -> str:
        return str(self.frames)

# -----------------------------------------------------------------------------
def _fill(color):
    return {"style": "filled", "fillcolor": color}

def _bfill(color):
    return {"style": "bold,filled", "fillcolor": color}


class FileSystemBacktrace(Backtrace):
    EDGES = dict(
        READ = r"^read$",
        WRITE = r"^write$",
        FSYNC = r"^fsync$",
        UNLINK = r"^unlink$",
        LSEEK = r"^lseek$",
        OPEN = r"^open$",
        CLOSE = r"^close$",
        READDIR = r"^readdir$",
        OPENDIR = r"^opendir$",
        MOUNT = r"^mount$",
        MKDIR = r"^mkdir$",
        STAT = r"^stat$",
        IRQ = r"^stm32_sdmmc_interrupt$",
        INIT = r"^stm32_sdio_initialize$",
    )
    COLORS = {**{"^fat_.*$": _fill("LightYellow"),
                 "^mmc?sd_.*$": _fill("LightCyan"),
                 "^stm32_.*$": _fill("DarkSeaGreen"),
                 "px4::logger::": _fill("Gold"),
                 "Navigator": _fill("GreenYellow"),
                 "can": _fill("LightSalmon"),
                 "^nsh_.*": _fill("LightSkyBlue"),
                 "^(param|bson)_.*": _fill("Bisque")},
              **{pattern: _bfill("Silver") for pattern in EDGES.values()}}


class SpiBacktrace(Backtrace):
    EDGES = dict(
        READ = r"RegisterRead|FIFOReadCount",
        WRITE = r"RegisterWrite",
        PARAM = r"^param_main$",
        NSH = r"^nsh_initialize$",
    )
    COLORS = {**{pattern: _bfill("Silver") for pattern in EDGES.values()},
              **{"^(ramtron|bch|bchlib)_.*$": _fill("LightYellow"),
                 "^spi_.*$": _fill("LightCyan"),
                 "^stm32_.*$": _fill("DarkSeaGreen"),
                 "ICM20602": _fill("Gold"),
                 "BMI088": _fill("GreenYellow"),
                 "ICM42688": _fill("LightSalmon"),
                 "^nsh_.*": _fill("LightSkyBlue"),
                 "^(param|bson)_.*": _fill("Bisque")}}


class I2cBacktrace(Backtrace):
    EDGES = dict(
        TRANSFER = r"stm32_i2c_transfer",
        NSH = r"^nsh_initialize$",
        IRQ = r"^stm32_i2c_isr$",
        INIT = r"^stm32_i2c_setclock$",
    )
    COLORS = {**{"^stm32_.*$": _fill("DarkSeaGreen"),
                 "BMM150": _fill("Yellow"),
                 "IST8310": _fill("Cyan"),
                 "BMP388": _fill("Gold"),
                 "RGBLED": _fill("LightSkyBlue"),
                 "MCP23009": _fill("LightSalmon")},
              **{pattern: _bfill("Silver") for pattern in EDGES.values()}}


class CanBacktrace(Backtrace):
    EDGES = dict(
        SEND = r"CanIface::send$",
        RECEIVE = r"CanIface::receive$",
        INIT = r"CanIface::init$",
        FRAME = r"TransferListener::handleFrame$",
    )
    COLORS = {**{pattern: _bfill("Silver") for pattern in EDGES.values()},
              **{"^(ramtron|bch|bchlib)_.*$": _fill("LightYellow"),
                 "Publisher": _fill("DarkSeaGreen")}}


class UartBacktrace(Backtrace):
    EDGES = dict(
        READ = r"^read$",
        WRITE = r"^write$",
        FFLUSH = r"^fflush$",
        FPUTC = r"^fputc$",
        FPUTS = r"^fputs$",
        OPEN = r"^open$",
        SYSLOG = r"^syslog$",
        PRINTF = r"^printf$",
        IRQ = r"^irq_dispatch$",
        INIT = r"^arm_earlyserialinit$",
    )
    COLORS = {**{"^I2CSPI": _fill("LightYellow"),
                 "^(up|uart)_.*$": _fill("LightCyan"),
                 "^stm32_.*$": _fill("DarkSeaGreen"),
                 "px4::logger": _fill("Gold"),
                 "[mM]avlink": _fill("GreenYellow"),
                 "can": _fill("LightSalmon"),
                 "^nsh_.*": _fill("LightSkyBlue"),
                 "^(param|bson)_.*": _fill("Bisque")},
              **{pattern: _bfill("Silver") for pattern in EDGES.values()}}


class SemaphoreBacktrace(Backtrace):
    # EDGES = dict(
    #     RESTORE = r"^nxsem_restoreholderprio$",
    #     BOOST =  r"^nxsem_boostholderprio$",
    #     WAIT = r"^nxsem_wait$",
    #     POST = r"^nxsem_post$",
    #     INIT = r"^sem_init$",
    #     DESTROY = r"^sem_destroy$",
    #     FREE = r"^free$",
    #     MALLOC = r"^malloc$",
    #     MEMSET = r"^memset$",
    # )
    # COLORS = {pattern: _bfill("Silver") for pattern in EDGES.values()}
    def __init__(self, description: str):
        """
        :param description: All lines of the backtrace command also containing a boost trace command.
        """
        super().__init__(description)
        ops = (_BoostOperation(d) for d in description.splitlines())
        ops = [o for o in ops if o.is_valid]
        self.operation = ops[0] if ops else None
        if self.operation is not None:
            self.type = self.operation.task
        if ops and self.operation.prio_to == 255:
            if self.type is not None:
                self.type += "_255"
            else:
                self.type = "255"



# Copyright (c) 2020-2022, Niklas Hauser
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: openocd.md
"""

from __future__ import annotations
import os
import time
import signal
import logging
import tempfile
import platform
import telnetlib
import subprocess
from pathlib import Path

from . import utils
from .backend import ProbeBackend

LOGGER = logging.getLogger("debug:openocd")

# -----------------------------------------------------------------------------
class OpenOcdBackend(ProbeBackend):
    """
    OpenOCD specific debug backend implementation. Starts `openocd` in
    its own subprocess and connects GDB to port `3333` of the localhost.

    See `call()` for additional information.
    """
    def __init__(self, commands: list[str] = None, config: list[Path] = None,
                 search: list[Path] = None):
        """
        :param commands: list of commands to execute on launch.
        :param config: list of configuration files to execute on launch.
        :param search: list of directories to search configuration files in.
        """
        super().__init__(":3333")
        self.commands = utils.listify(commands)
        self.config = utils.listify(config)
        self.search = utils.listify(search)
        self.process = None
        self.name = "openocd"

    def start(self):
        self.process = call(self.commands, self.config, self.search,
                            blocking=False, silent=True)
        LOGGER.info(f"Starting {self.process.pid}...")

    def stop(self):
        if self.process is not None:
            LOGGER.info(f"Stopping {self.process.pid}.")
            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
            os.waitpid(os.getpgid(self.process.pid), 0)
            self.process = None


def call(commands: list[str] = None, config: list[Path] = None,
         search: list[Path] = None, blocking: bool = True,
         silent: bool = False)  -> "int | subprocess.Popen":
    """
    Starts `openocd` and connects to the microcontroller without resetting the
    device. You can overwrite the default binary by exporting an alternative in
    your environment:

    ```sh
    export PX4_OPENOCD=path/to/openocd
    ```

    :param commands: list of commands to execute on launch.
    :param config: list of configuration files to execute on launch.
    :param search: list of directories to search configuration files in.
    :param blocking: Run in current process as a blocking call.
                     Set to `False` to run in a new subprocess.
    :param silent: Disable any reporting from `openocd`.

    :return: The process return code if `blocking` or the Popen object.
    """
    commands = ["init"] + utils.listify(commands)
    config = utils.listify(config)
    search = utils.listify(search)
    if silent:
        null_file = "/dev/null"
        commands.append("log_output " + null_file)

    # Provide additional search paths via the OPENOCD_SCRIPTS environment variable
    # See http://openocd.org/doc/html/Running.html
    # os.environ.get("OPENOCD_SCRIPTS", "")

    binary = os.environ.get("PX4_OPENOCD", "openocd")

    command_openocd = "{} {} {} {}".format(
        binary,
        " ".join(map('-s "{}"'.format, search)),
        " ".join(map('-f "{}"'.format, config)),
        " ".join(map('-c "{}"'.format, commands))
    )
    LOGGER.debug(command_openocd)

    kwargs = {"cwd": os.getcwd(), "shell": True}
    if blocking:
        return subprocess.call(command_openocd, **kwargs)

    # We have to start openocd in its own session ID, so that Ctrl-C in GDB
    # does not kill OpenOCD.
    kwargs["start_new_session"] = True
    return subprocess.Popen(command_openocd, **kwargs)


# -----------------------------------------------------------------------------
def itm(backend, fcpu: int, baudrate: int = None) -> int:
    """
    Launches `openocd` and configured ITM output on the SWO pin.
    The data is written into a temporary file, which is `tail`'ed for display.

    :param backend: A OpenOCD backend.
    :param fcpu: CPU frequency of the target.
    :param baudrate: optional frequency of the SWO connection.

    :return: the process return code
    """
    if not fcpu:
        raise ValueError("fcpu must be the CPU/HCLK frequency!")

    with tempfile.NamedTemporaryFile() as tmpfile:
        backend.commands += [
            "tpiu create itm.tpiu -dap [dap names] -ap-num 0 -protocol uart",
           f"itm.tpiu configure -traceclk {fcpu} -pin-freq {baudrate or 2000000} -output {tmpfile}",
            "itm.tpiu enable",
            "tpiu init",
            "itm port 0 on",
        ]
        # Start OpenOCD in the background
        with backend.scope():
            # Start a blocking call to monitor the log file
            # TODO: yield out new log lines in the future
            try:
                subprocess.call("tail -f {}".format(tmpfile.name),
                                cwd=os.getcwd(), shell=True)
            except KeyboardInterrupt:
                pass
    return 0

def rtt(backend, channel: int = 0) -> int:
    """
    Launches the backend in the background and connects a telnet client to the
    Real-Time Transfer process. You can disconnect with Ctrl+D.

    :param backend: A OpenOCD backend object.
    :param channel: The RTT channel to connect to.

    :return: the process return code
    """
    backend.commands += [
        "rtt setup 0x20000000 256000",
        "rtt start",
        "rtt polling_interval 1",
       f"rtt server start {9090 + channel} {channel}",
    ]
    import telnetlib
    # Start OpenOCD in the background
    with backend.scope():
        with telnetlib.Telnet("localhost", 9090+channel) as tn:
            try:
                tn.interact()
            except KeyboardInterrupt:
                pass
    return 0


# -----------------------------------------------------------------------------
def program(source: Path, commands: list[str] = None, config: list[Path] = None,
            search: list[Path] = None) -> int:
    """
    Loads the source file into the microcontroller and resets the device.

    :param commands: list of commands to execute on launch.
    :param config: list of configuration files to execute on launch.
    :param search: list of directories to search configuration files in.
    :param source: path to a `.elf` file to upload.

    :return: the process return code of openocd
    """
    from .gdb import call as gdb_call
    backend = OpenOcdBackend(commands, config, search)
    gdb_cmds = ["monitor reset halt", "load", "monitor reset run", "quit"]
    return gdb_call(backend, source, ui="batch", commands=gdb_cmds)
    # Unfortunately, the OpenOCD program command seems to erase Flash sector 0
    # even if the ELF file has an offset. This overwrites the bootloader and
    # bricks the FMU, so we use GDB instead.
    # commands = utils.listify(commands) + \
    #     [f"program {Path(source).absolute()} preverify verify reset exit"]
    # return call(commands=commands, config=config, search=search)


def reset(commands: list[str] = None, config: list[Path] = None,
          search: list[Path] = None) -> int:
    """
    Resets the device via OpenOCD.

    :param commands: list of commands to execute on launch.
    :param config: list of configuration files to execute on launch.
    :param search: list of directories to search configuration files in.

    :return: the process return code of OpenOCD
    """
    commands = utils.listify(commands) + ["reset", "shutdown"]
    return call(commands=commands, config=config, search=search)


# -----------------------------------------------------------------------------
def _add_subparser(subparser):
    parser = subparser.add_parser("openocd", help="Use OpenOCD as Backend.")
    parser.add_argument(
        "-f",
        dest="oconfig",
        action="append",
        help="Use these OpenOCD config files.")
    parser.add_argument(
        "-s",
        dest="osearch",
        action="append",
        help="Search in these paths for config files.")
    parser.add_argument(
        "-c",
        dest="ocommands",
        action="append",
        help="Extra OpenOCD commands.")
    parser.add_argument(
        "--speed",
        dest="ospeed",
        type=int,
        default=8000,
        choices=[24000, 8000, 3300, 1000, 200, 50, 5],
        help="SWD baudrate in kHz.")
    parser.set_defaults(backend=lambda args:
        OpenOcdBackend([f"adapter speed {args.ospeed}"] +
                       utils.listify(args.ocommands),
                       args.oconfig, args.osearch))
    return parser


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse
    import emdbg

    parser = argparse.ArgumentParser(
        description="OpenOCD debug probe: Upload, reset, and logging")
    parser.add_argument(
        "-f",
        dest="config",
        action="append",
        help="Use these OpenOCD config files.")
    parser.add_argument(
        "-s",
        dest="searchdirs",
        action="append",
        help="Search in these paths for config files.")
    parser.add_argument(
        "-c",
        dest="commands",
        action="append",
        help="Extra OpenOCD commands.")
    parser.add_argument(
        "--speed",
        type=int,
        default=8000,
        choices=[24000, 8000, 3300, 1000, 200, 50, 5],
        help="SWD baudrate in kHz.")
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        help="Verbosity level.")

    subparsers = parser.add_subparsers(title="Command", dest="command")

    subparsers.add_parser("reset", help="Reset the device.")

    subparsers.add_parser("run", help="Run OpenOCD.")

    upload_parser = subparsers.add_parser("upload", help="Upload firmware.")
    upload_parser.add_argument(
        "--source",
        required=True,
        help="The firmware to upload: `.elf` file")

    args = parser.parse_args()
    emdbg.logger.configure(args.verbosity)

    commands = [f"adapter speed {args.speed}"] + utils.listify(args.commands)
    if not (config := args.config):
        config = ["interface/stlink.cfg", "target/stm32f7x.cfg"]

    if args.command == "reset":
        exit(reset(commands, config, args.searchdirs))

    if args.command == "run":
        exit(call(commands, config, args.searchdirs, blocking=True))

    if args.command == "upload":
        exit(program(args.source, commands, config, args.searchdirs))

    LOGGER.error("Unknown command!")
    exit(1)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
PX4 generates a log file if it encounters a hardfault.
This tool converts this to a coredump that can be loaded with CrashDebug.
To interpret the hardfault location and reason, use the GDB plugins.

.. note::
   The hardfault log only contains a copy of the stack and registers. The
   converter fills unknown memory with the `0xfeedc0de` value, which may then
   display wrong or incomplete results in some of the GDB plugins. Keep this in
   mind when converting a hardfault to a coredump.


## Command Line Interface

To convert the hardfault log:

```sh
# convert to hardfault_coredump.txt
python3 -m emdbg.analyze.hardfault hardfault.log
# or with an explicit name
python3 -m emdbg.analyze.hardfault hardfault.log -o custom_name.txt
```
"""

from __future__ import annotations
import re
from pathlib import Path
import statistics
import itertools
from collections import defaultdict

# FIXME: hardcoded for FMUv6x memory layout (also works for FMUv5x)
_ADDR_RANGES = [
    (0x2000_0000, 0x80000),
    (0x2400_0000, 0x80000),
    (0x3000_0000, 0x48000),
    (0x3800_0000, 0x10000),
    (0x3880_0000, 0x01000),
    (0xE004_2000, 4), # DEVID F7
    (0x5C00_1000, 4), # DEVID H7
    (0xE000_E000, 0xFFF), # SCS
]

_UNKNOWN_MEM = 0xfeedc0de

_UNKNOWN_REGS = [
    "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11", "r12", "sp", "lr", "pc",
    "xpsr", "msp", "psp", "primask", "basepri", "faultmask", "control", "fpscr",
    "s0", "s1", "s2", "s3", "s4", "s5", "s6", "s7", "s8", "s9", "s10", "s11", "s12", "s13", "s14", "s15", "s16",
    "s17", "s18", "s19", "s20", "s21", "s22", "s23", "s24", "s25", "s26", "s27", "s28", "s29", "s30", "s31",
    "d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7", "d8", "d9", "d10", "d11", "d12", "d13", "d14", "d15",
]

def convert(log: str) -> str:
    """
    Convert a hardfault log to

    :param log: The content of a hardfault log file.
    :return: A formatted string containing the coredump.
    """
    lines = log.splitlines()

    known_mems = {}
    regs = {r: _UNKNOWN_MEM for r in _UNKNOWN_REGS}
    fault_regs = {}
    for line in lines:
        if len(reg_matches := re.findall(r" (r\d+|sp|lr|pc|xpsr|basepri|control|primask):\[?(0x[0-9a-f]+)]?", line)):
            for reg, val in reg_matches:
                regs[reg] = int(val, 16)
                if reg == "sp":
                    regs["msp"] = regs["sp"]
                    regs["psp"] = regs["sp"]

        if len(fault_matches := re.findall(r" (c|h|d|mm|b|a|ab)fsr:\[?(0x[0-9a-f]+)]?", line)):
            for reg, val in fault_matches:
                fault_regs[reg] = int(val, 16)

        m_base = re.match(r"^(0x[0-9a-f]+)\s(0x[0-9a-f]+).*$", line)
        m_correlated = re.match(r"^(0x[0-9a-f]+)\s0x[0-9a-f]+ -> \[(0x[0-9a-f]+)].*$", line)

        m = m_correlated or m_base

        if m:
            addr = int(m.group(1), 16)
            val = int(m.group(2), 16)
            known_mems[addr] = val

    # FIXME: hardcoded for FMUv5x (STM32F765)
    known_mems[0xE004_2000] = 0x10030451
    # FIXME: hardcoded for FMUv6x (STM32H753)
    known_mems[0x5C00_1000] = 0x10030450

    # CPUID
    known_mems[0xE000_ED00] = 0x411fc270
    # dump all the fault registers back into memory
    known_mems[0xE000_ED28] = fault_regs.get("c", 0)
    known_mems[0xE000_ED2C] = fault_regs.get("h", 0)
    known_mems[0xE000_ED30] = fault_regs.get("d", 0)
    known_mems[0xE000_ED34] = fault_regs.get("mm", 0)
    known_mems[0xE000_ED38] = fault_regs.get("b", 0)
    known_mems[0xE000_ED3C] = fault_regs.get("a", 0)
    known_mems[0xE000_EFA8] = fault_regs.get("ab", 0)

    output = []
    for mem_start, mem_size in _ADDR_RANGES:
        for addr in range(mem_start, mem_start + mem_size, 16):
            fmt = "{:#8x}:\t{:#8x}\t{:#8x}\t{:#8x}\t{:#8x}"
            fmt = fmt.format(addr, known_mems.get(addr, _UNKNOWN_MEM),
                                   known_mems.get(addr + 4, _UNKNOWN_MEM),
                                   known_mems.get(addr + 8, _UNKNOWN_MEM),
                                   known_mems.get(addr + 12, _UNKNOWN_MEM))
            output.append(fmt)

    for reg, val in regs.items():
        output.append(f"{reg:15} {val:#20x} {val:20}")

    return "\n".join(output)



# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Hardfault log converter")
    parser.add_argument(
        "log",
        type=Path,
        help="The hardfault log.")
    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        default=None,
        help="The GDB log containing the semaphore boost trace.")
    args = parser.parse_args()

    if (outfile := args.output) is None:
        outfile = args.log.with_suffix("_coredump.txt")

    outfile.write_text(convert(args.log.read_text()))







# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# Remote GDB Access

This module provides access to remotely connected GDB and allows partial or
complete control over the GDB command prompt or the entire GDB Python API.

Currently two access methods are available:

- `emdbg.debug.remote.mi.Gdb`:
		command prompt access via the GDB/MI protocol created with
		`emdbg.debug.gdb.call_mi()`.
- `emdbg.debug.remote.rpyc.Gdb`:
		Full Python API access via RPyC created with
		`emdbg.debug.gdb.call_rpyc()`.
"""

__all__ = [
	"gdb",
	"rpyc",
	"mi",
]

from . import rpyc
from . import mi



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# GDB Python Modules

This module includes many PX4-specific GDB Python plugins that provide additional
debug functionality for interactive debugging and automated scripting.

Since we want these modules to work inside and outside of GDB, we do not use
`import gdb` which is only available inside GDB, but instead provide `gdb`
as a function argument so that it can be replaced by
`emdbg.debug.remote.rpyc.Gdb`.

.. warning::
    You must only use the [GDB Python API][api] and the Python standard library
    in this folder so that they work both inside and outside of GDB.


[api]: https://sourceware.org/gdb/onlinedocs/gdb/Python-API.html
"""



from .task import all_tasks, all_tasks_as_table, all_files_as_table, task_switch
from .semaphore import Semaphore
from .buffer import UartBuffer
from .device import all_registers, all_registers_as_table, all_gpios_as_table
from .device import vector_table, vector_table_as_table, Device, coredump
from .device import discover as discover_device
from .svd import PeripheralWatcher

from .system_load import restart_system_load_monitor
from .utils import gdb_backtrace as backtrace
from .data import pinout


_TARGET = None
_SVD_FILE = None



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: fmu.md
"""

from __future__ import annotations
import time
from functools import cached_property
from pathlib import Path
from contextlib import contextmanager, nullcontext
import emdbg


class Fmu:
    """
    FMU test bench with optional nsh, power, and logic analyzer attachments.
    """
    def _DBGMCU_CONFIG(target):
        # Vector catch all exceptions, but not reset
        common = ["set *0xE000EDFC = *0xE000EDFC | 0x07f0"]
        # Halt all timers and peripherals while debugging
        if "fmu-v5x" in target:
            return ["set *0xE0042008 = 0xffffffff",
                    "set *0xE004200C = 0xffffffff"] + common
        if "fmu-v6x" in target:
            return ["set *0xE00E1034 = 0xffffffff",
                    "set *0xE00E103C = 0xffffffff",
                    "set *0xE00E104C = 0xffffffff",
                    "set *0xE00E1054 = 0xffffffff"] + common
        return []

    def __init__(self, target: str, elf: Path,
                 gdb: "emdbg.debug.remote.gdb.Interface",
                 nsh: "emdbg.serial.protocol.Nsh" = None,
                 power: "emdbg.power.base.Base" = None,
                 io: "emdbg.io.digilent.Digilent" = None):
        self.elf: Path = elf
        """The ELF file under debug"""
        self.gdb: "emdbg.debug.remote.gdb.Interface" = gdb
        """The remote GDB access interface object"""
        self.nsh: "emdbg.serial.protocol.Nsh" = nsh
        """The NSH protocol controller"""
        self.power: "emdbg.power.base.Base" = power
        """The power relay controlling the FMU"""
        self.io: "emdbg.io.digilent.Digilent" = io
        """The Digilent Scope"""
        self._target = target

    def _init(self):
        self.gdb.interrupt_and_wait()
        for cmd in self._DBGMCU_CONFIG(self._target):
            self.gdb.execute(cmd)
        self.restart_system_load_monitor()

    def _deinit(self):
        self.gdb.interrupt_and_wait()
        self.gdb.execute("disconnect")
        self.gdb.interrupted = False
        # Give JLink some time to react to the disconnect before shut down
        time.sleep(1)

    def restart_system_load_monitor(self):
        """See `emdbg.debug.px4.system_load.restart_system_load_monitor`"""
        with self.gdb.interrupt_continue():
            if self.gdb.type == "mi":
                self.gdb.execute("python px4.restart_system_load_monitor(gdb)")
            else:
                emdbg.debug.px4.restart_system_load_monitor(self.gdb)

    def coredump(self, filename: Path = None):
        """
        Dumps the FMU core for later analysis with CrashDebug (see
        `emdbg.debug.crashdebug`).

        :param filename: Default `coredump_{datetime}.txt`.
        """
        with self.gdb.interrupt_continue():
            if False:
                # Connection is remote, therefore we must use slower RPyC interface
                emdbg.debug.px4.coredump(self.gdb, filename=filename)
            else:
                # Executing directly on the GDB process is *significantly* faster!
                if filename: filename = f"--file '{filename}'"
                self.gdb.execute(f"px4_coredump {filename or ''}")

    def upload(self, source: Path = None):
        """
        Uploads the ELF file to the FMU, resets the device, clears the NSH
        :param source: optional path to ELF file, default is the ELF file passed
                       to the constructor.
        """
        with self.gdb.interrupt_continue():
            self.gdb.execute("monitor reset")
            self.gdb.execute(f"load {source or self.elf}", timeout=60)
            self.reset()
            if self.nsh is not None:
                self.nsh.clear()

    def power_on_reset(self, delay_off: float = 1, delay_on: float = 2):
        """
        Disconnects GDB, stops the backend, power cycles the entire bench,
        restarts the backend, reconnects and reinitializes GDB.
        **Only works if initialized with a power relay!**
        """
        if self.power is None:
            LOGGER.warning("Power Relay not configured!")
            return
        # Disconnect and stop the backend
        self._deinit()
        self.gdb.backend.stop()
        # Wait for JLink to shut down, os.waitpid is not enough
        time.sleep(1)

        # Have you tried turning it off and on again?
        self.power.cycle(delay_off, delay_on)

        # Restart and reattach the backend
        self.gdb.backend.start()
        # Wait for JLink to start up again, since it starts non-blocking
        time.sleep(2)
        # Reinitialize GDB to attach back to JLink
        for cmd in self.gdb.backend.init(self.elf):
            self.gdb.execute(cmd)
        self._init()

    def reset(self):
        """Resets the FMU"""
        with self.gdb.interrupt_continue():
            self.gdb.execute("monitor reset")
            self._init()

    def sleep(self, seconds: float, func=None):
        """
        Sleeps and calls `func` every second. This can be useful for polling if
        a condition has happened.
        """
        while(seconds > 0):
            seconds -= 1
            time.sleep(1)
            if func is not None:
                func(seconds)


def _px4_config(px4_directory: Path, target: Path, commands: list[str] = None,
                ui: str = None, speed: int = 16000, backend: str = None) -> tuple:
    if "fmu-v5x" in target:
        device = "STM32F765II"
        config = "fmu_v5x"
    elif "fmu-v6x" in target:
        device = "STM32H753II"
        config = "fmu_v6x"
    else:
        raise ValueError(f"Unknown device for '{target}'!")

    px4_dir = Path(px4_directory).absolute().resolve()
    data_dir = Path(__file__).parent.resolve() / "data"

    if backend in ["stlink", "orbtrace"]:
        config += f"_{backend}.cfg"
        backend_obj = emdbg.debug.OpenOcdBackend(config=[data_dir / config])
    elif backend == "jlink":
        rtos_so = px4_dir / f"platforms/nuttx/NuttX/nuttx/tools/jlink-nuttx.so"
        if not rtos_so.exists(): rtos_so = None
        backend_obj = emdbg.debug.JLinkBackend(device, speed, rtos_so)
    elif isinstance(backend, Path):
        backend_obj = emdbg.debug.CrashProbeBackend(backend)
    elif ":" in backend:
        backend_obj = emdbg.debug.ProbeBackend(backend)
    else:
        raise ValueError(f"Unknown backend '{backend}'!")

    if Path(target).suffix == ".elf":
        elf = Path(target)
    else:
        elf = next((px4_dir / "build" / target).glob("*.elf"), None)
        if elf is None:
            raise ValueError(f"Cannot find ELF file in build folder '{px4_dir}/build/{target}'!")

    cmds = [f"dir {px4_dir}",
            f"source {data_dir}/fmu.gdb",
            f"python px4._TARGET='{target.lower()}'"]
    if ui is not None:
        cmds += Fmu._DBGMCU_CONFIG(target)
    cmds += (commands or [])

    return backend_obj, elf, cmds


# -----------------------------------------------------------------------------
@contextmanager
def debug(px4_directory: Path, target: str, serial: str = None,
          digilent: str = None, power: "emdbg.power.base.Base" = None,
          ui: str = None, commands: list[str] = None, with_rpyc: bool = False,
          keep_power_on: bool = True, upload: bool = True, backend: str = None,
          _FmuClass = Fmu) -> Fmu:
    """
    Launches and configures the Fmu test bench.
    1. Switches on the power relay if specified.
    2. Starts the J-Link GDB server in the background.
    3. Launches GDB and connects it to the debug probe backend.
    4. Connects and initializes the NSH if specified.
    5. Connects and initializes the Digilent Scope if specified.
    6. Loads the PX4 specific GDB scripts, the `emdbg.debug.px4`
       modules Python modules, and the `fmu.gdb` script.
    7. Uploads the latest firmware to the FMU.
    8. Yields the test bench in halted reset state.

    :param px4_directory: path to the PX4-Autopilot repository you want to debug.
    :param target: target name as a string, for example, `px4_fmu-v5x`.
                   Can also be a path to an ELF file.
    :param serial: optional serial number of the USB-Serial bridge for the NSH.
                   If `None`, then `Fmu.nsh` is empty and cannot be used.
    :param digilent: optional serial number of the Digilent. If `None`, then
                     `Fmu.io` is empty and cannot be used.
    :param ui: If not `None`, then this launches the interactive debugger via
               `emdbg.debug.gdb.call` instead of the scripting API.
    :param commands: list of additional GDB commands to execute during launch.
    :param with_rpyc: Use the RPyC GDB interface implementation, rather than the
                      GDB/MI interface. See `emdbg.debug.remote`.
    :param keep_power_on: Do not shut off the Fmu after the context has closed.
    :param upload: Automatically upload the firmware after powering on the FMU.
    :param backend: `openocd`, `jlink`, or `IP:PORT` for a GDB server on another machine.

    :return: A configured test bench with the latest firmware.
    """
    backend, elf, cmds = _px4_config(px4_directory, target, commands, ui,
                                     backend=backend or "openocd")

    with (nullcontext() if power is None else power) as pwr:
        try:
            if ui is not None:
                # Manual mode that only connects the debugger (blocking)
                if power: pwr.on()
                yield emdbg.debug.gdb.call(backend, elf, commands=cmds, ui=ui)
            else:
                # Turn off, then connect the serial
                if power: pwr.off()
                serial = emdbg.serial.nsh(serial) if serial is not None else nullcontext()
                scope = emdbg.io.analog_discovery(digilent) if digilent is not None else nullcontext()
                with serial as nsh, scope as io:
                    # Then power on and connect GDB to get the full boot log
                    if power: pwr.on()
                    gdb_call = emdbg.debug.gdb.call_rpyc if with_rpyc else emdbg.debug.gdb.call_mi
                    debugger = gdb_call(backend, elf, commands=cmds)
                    with debugger as gdb:
                        bench = _FmuClass(target, elf, gdb, nsh, pwr, io)
                        if upload: bench.upload()
                        yield bench
                        bench._deinit()
        finally:
            if not keep_power_on:
                pwr.off(delay=0)


# -----------------------------------------------------------------------------
@contextmanager
def shell(serial: str = None, power = None, keep_power_on: bool = True) -> Fmu:
    """
    Launches and configures the Fmu test bench.
    1. Switches on the power relay if specified.
    4. Connects and initializes the NSH if specified.
    8. Yields the NSH after reset state.

    :param serial: optional serial number of the USB-Serial bridge for the NSH.
                   If `None`, then `Fmu.nsh` is empty and cannot be used.
    :param keep_power_on: Do not shut off the Fmu after the context has closed.

    :return: A configured test bench with the latest firmware.
    """

    if power is None: power = nullcontext()
    with power as pwr:
        try:
            # Turn off, then connect the serial
            pwr.off()
            serial = emdbg.serial.nsh(serial) if serial is not None else nullcontext()
            with serial as nsh:
                # Then power on and connect GDB to get the full boot log
                pwr.on()
                yield nsh
        finally:
            if not keep_power_on:
                pwr.off(delay=0)

# -----------------------------------------------------------------------------
def _arguments(description, modifier=None):
    import argparse, emdbg
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--px4-dir",
        default=".",
        type=Path,
        help="The PX4 root directory you are working on.")
    parser.add_argument(
        "--target",
        default="px4_fmu-v5x",
        help="The target you want to debug.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--jlink",
        default=False,
        action="store_true",
        help="Use a J-Link debug probe")
    group.add_argument(
        "--stlink",
        default=False,
        action="store_true",
        help="Use an STLink debug probe")
    group.add_argument(
        "--orbtrace",
        default=False,
        action="store_true",
        help="Use an ORBtrace mini debug probe")
    group.add_argument(
        "--remote",
        help="Connect to a remote GDB server: 'IP:PORT'")
    group.add_argument(
        "--coredump",
        type=Path,
        help="Inspect a GDB coredump or PX4 hardfault log.")
    parser.add_argument(
        "--ui",
        default="cmd",
        choices=["tui", "gdbgui", "cmd", "batch"],
        help="The user interface you want to use.")
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Verbosity level.")
    parser.add_argument(
        "-ex",
        dest="commands",
        action="append",
        help="Extra GDB commands.")
    if modifier: modifier(parser)
    args = parser.parse_args()
    emdbg.logger.configure(args.verbosity)
    backend = args.remote
    if args.stlink: backend = "stlink"
    if args.jlink: backend = "jlink"
    if args.orbtrace: backend = "orbtrace"
    if args.coredump: backend = args.coredump
    return args, backend


if __name__ == "__main__":
    args, backend = _arguments("Debug FMU")

    with debug(args.px4_dir, args.target, ui=args.ui, commands=args.commands,
               backend=backend) as gdb_call:
        exit(gdb_call)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from pathlib import Path
from .operation import CopyOperation, PatchOperation, PatchManager

def _data(file: str) -> Path:
    return Path(__file__).parent / "data" / file


def semaphore_boostlog(px4_root: Path) -> PatchManager:
    """
    Enable runtime logging of task priority boosting through semaphores.
    This adds a lock-free 32 slot buffer containing task and priority information
    that is written from inside the kernel and read in the PX4 logger module,
    which logs it out to the NSH.
    """
    px4_root = Path(px4_root)
    operations = [
        PatchOperation(px4_root, _data("semaphore_boostlog.patch")),
        CopyOperation(_data("sem_boostlog.c"),
                      px4_root / "platforms/nuttx/NuttX/nuttx/sched/semaphore/sem_boostlog.c"),
        CopyOperation(_data("semaphore_boostlog.h"),
                      px4_root / "platforms/nuttx/NuttX/nuttx/include/semaphore_boostlog.h"),
    ]
    return PatchManager("Logging of Task Priority Inheritance by Semaphores", operations)


def reduce_firmware_size_v5x(px4_root: Path) -> PatchManager:
    """
    Disables UAVCAN and a few drivers to make the binary size fit on the FMUv5x.
    """
    px4_root = Path(px4_root)
    operations = [
        PatchOperation(px4_root, _data("disable_uavcan_v5x.patch")),
    ]
    return PatchManager("Make the Firmware fit on the FMUv5x Flash", operations)


def nuttx_tracing_itm(px4_root: Path) -> PatchManager:
    """
    Adds scheduler and heap instrumentation to NuttX via ITM.
    Requires the `itm_logging` patch.
    """
    px4_root = Path(px4_root)
    operations = [
        CopyOperation(_data("itm.h"),
                      px4_root / "platforms/nuttx/NuttX/nuttx/include/nuttx/itm/itm.h"),
        CopyOperation(_data("itm_Make.defs"),
                      px4_root / "platforms/nuttx/NuttX/nuttx/drivers/itm/Make.defs"),
        PatchOperation(px4_root, _data("itm_nuttx_Makefile.patch")),
        PatchOperation(px4_root, _data("nuttx_tracing_itm.patch")),
    ]
    return PatchManager("Add tracing support to NuttX via ITM", operations)


def nuttx_sdmmc_reg_access(px4_root: Path) -> PatchManager:
    """
    Un-inlines the `sdmmc_putreg32`, `sdmmc_getreg32`, and `sdmmc_modifyreg32`
    functions, so that breakpoints can be set on them.
    """
    px4_root = Path(px4_root)
    operations = [
        PatchOperation(px4_root, _data("sdmmc_no_inline.patch")),
    ]
    return PatchManager("Un-inline SDMMC register access", operations)


def malloc_return_null(px4_root: Path) -> PatchManager:
    """
    Instruments the mm_malloc in NuttX to fail if call count equals global
    `emdbg_malloc_count_null` variable, which is placed in `.noinit` section.
    You can use this to find code that doesn't check the malloc return value for
    NULL or see how the error handling performs.
    """
    px4_root = Path(px4_root)
    operations = [
        PatchOperation(px4_root, _data("malloc_return_null.patch")),
    ]
    return PatchManager("Make the n-th malloc call return NULL", operations)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: calltrace.md
"""

if __name__ == "__main__":
    import re
    import emdbg
    import argparse
    from pathlib import Path
    from .utils import read_gdb_log
    from ..bench.fmu import _arguments

    values = {
        "FileSystem": emdbg.analyze.FileSystemBacktrace,
        "SPI": emdbg.analyze.SpiBacktrace,
        "I2C": emdbg.analyze.I2cBacktrace,
        "CAN": emdbg.analyze.CanBacktrace,
        "UART": emdbg.analyze.UartBacktrace,
        "Semaphore": emdbg.analyze.SemaphoreBacktrace,
        "Generic": emdbg.analyze.Backtrace,
    }
    def _modifier(parser):
        parser.add_argument(
            "--log-prefix",
            type=Path,
            default="calltrace",
            help="Log file name.")
        parser.add_argument(
            "--sample",
            default=60,
            type=int,
            help="Sample time in seconds.")
        parser.add_argument(
            "--type",
            choices=values.keys(),
            default="Generic",
            help="The backtrace class to use.")

    args, backend = _arguments("Generate call graphs of a function or memory location", _modifier)
    print(f"Logging for: {', '.join(args.commands)}")

    calltrace = "".join(filter(lambda c: re.match(r"[\w\d]", c), '_'.join(args.commands)))
    calltrace = Path(f"{args.log_prefix}_{calltrace}.txt")
    with emdbg.bench.fmu(args.px4_dir, args.target, backend=backend, upload=False) as bench:
        for ex in args.commands:
            bench.gdb.execute(ex)
            if any(ex.startswith(b) for b in ["break ", "watch ", "awatch ", "rwatch "]):
                bench.gdb.execute("px4_commands_backtrace")
        bench.gdb.execute(f"px4_log_start {calltrace}")
        bench.gdb.continue_nowait()

        bench.sleep(args.sample)

        # halt the debugger and stop logging
        bench.gdb.interrupt_and_wait()
        bench.gdb.execute("px4_log_stop")

        backtraces = re.split(r"(?:Breakpoint|Hardware .*?watchpoint) \d", read_gdb_log(calltrace)[20:])
        emdbg.analyze.callgraph_from_backtrace(backtraces, values.get(args.type),
                                               output_graphviz=calltrace.with_suffix(".svg"))
